/* Copyright (C) 2019-2022 Marvell */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>
#include <dirent.h>
#include <signal.h>
#include <errno.h>

#ifdef AUTH_ELS_KMIP	// If we are not compiling KMIP, then don't want these
#include <openssl/bio.h>
#endif

#include <collections/linked_list.h>
#include <threading/thread.h>
#include <threading/condvar.h>
#include <threading/mutex.h>
#include <processing/jobs/callback_job.h>
#include <credentials/keys/shared_key.h>
#include <utils/identification.h>
#include <collections/hashtable.h>

#include <config/child_cfg.h>

#include "auth_els_configs.h"
#include "auth_els_utils.h"

#define DEFAULT_REAUTH_REKEY_JITTER 30 // minimum time between reauth and rekey
#define CONFIG_LIMIT_OVERRIDE_LIMITS_COOKIE 18652001

enum shared_secret_type_t {
	SHARED_SECRET_TYPE_WWPN,
	SHARED_SECRET_TYPE_CONFIG_FILE,
	SHARED_SECRET_TYPE_KMIP,
};

struct private_auth_els_configs_t {

	auth_els_configs_t public;

	unsigned int reauth_rekey_jitter;
	bool override_limits;				// Allows overriding of minimum reauth_time and rekey_time
	
	chunk_t shared_secret;
	enum shared_secret_type_t secret_type;
	
	proposal_t *esp_proposal;
	proposal_t *ike_proposal;
	bool vmware;
};
typedef struct private_auth_els_configs_t private_auth_els_configs_t;

METHOD(auth_els_configs_t, reload_config, bool,
	private_auth_els_configs_t *this)
{
	char *remote, *wwpn, *disabled_remote;
	int wwpn_len = 0, index = 0, j = 0;
	int cookie;
	char *rekey;

	enumerator_t* enumerator = lib->settings->create_section_enumerator(lib->settings,
									"%s.plugins.auth-els.disabled_remotes", lib->ns);

	while (enumerator->enumerate(enumerator, &remote))
	{
		DBG2(DBG_CFG, "disabled remotes: %s", remote);
		wwpn = lib->settings->get_str(lib->settings,
							"%s.plugins.auth-els.disabled_remotes.%s.wwpn", NULL,
							lib->ns, remote);
		if (wwpn == NULL)
		{
			DBG2(DBG_CFG, "disabled remotes NOT found in auth-els.conf");
			continue;
		}
		DBG2(DBG_CFG, "disabled remote port for auth from conf '%s'", wwpn);
		wwpn_len = strlen(wwpn) + 1;
		disabled_remote = (char*)malloc(wwpn_len);
		memset(disabled_remote, 0, wwpn_len);
		for (index = 0, j = 0; index < strlen(wwpn); index++)
		{
			if (wwpn[index] == ':')
				continue;
			if (wwpn[index] == '-')
				continue;
			if (wwpn[index] == '.')
				continue;
			if (wwpn[index] == ' ')
				continue;

			disabled_remote[j++] = wwpn[index];
			if (index == (strlen(wwpn) - 1))
				disabled_remote[j] = '\0';
		}


		if(j > 0)
		{

		}
		else
		{
			free (disabled_remote);
		}
	}
	enumerator->destroy(enumerator);

	enumerator = lib->settings->create_section_enumerator(lib->settings,
			"%s.plugins.auth-els.rekey_config", lib->ns);

	while (enumerator->enumerate(enumerator, &rekey))
	{
		this->reauth_rekey_jitter = lib->settings->get_int(lib->settings,
				"%s.plugins.auth-els.rekey_config.%s.reauth_rekey_jitter", DEFAULT_REAUTH_REKEY_JITTER,
				lib->ns, rekey);
		
		cookie = lib->settings->get_int(lib->settings,
				"%s.plugins.auth-els.rekey_config.%s.override_lower_limits", 0,
				lib->ns, rekey);
		
		if (cookie == CONFIG_LIMIT_OVERRIDE_LIMITS_COOKIE)
		{
			this->override_limits = true;
		}
	}
	enumerator->destroy(enumerator);

	int sender_thread_count = lib->settings->get_int(lib->settings,
									"%s.sender_thread_count", 1, lib->ns);
	
	this->vmware = lib->settings->get_bool(lib->settings,
					"%s.plugins.auth-els.vmware", FALSE, lib->ns);

	DBG_STD ("config setting of reauth_rekey_jitter: %d, override_limits: %d, sender_thread_count: %d",
					this->reauth_rekey_jitter, this->override_limits, sender_thread_count);
	
	return TRUE;
}

METHOD(auth_els_configs_t, get_vmware, bool,
	private_auth_els_configs_t *this)
{
	return this->vmware;
}

METHOD(auth_els_configs_t, get_reauth_rekey_jitter, unsigned int,
	private_auth_els_configs_t *this)
{
	return this->reauth_rekey_jitter;
}

METHOD(auth_els_configs_t, get_override_limits, bool,
	private_auth_els_configs_t *this)
{
	return this->override_limits;
}

METHOD(auth_els_configs_t, destroy, void,
	private_auth_els_configs_t *this)
{
	DBG_ENTER;
	
	free(this);
}

auth_els_configs_t *auth_els_configs_create()
{
	private_auth_els_configs_t *this;

	INIT(this,
		.public = {
			.destroy = _destroy,
			.reload_config = _reload_config,
			.get_reauth_rekey_jitter = _get_reauth_rekey_jitter,
			.get_override_limits = _get_override_limits,
			.get_vmware = _get_vmware,
		},
		.override_limits = false,
		.vmware = false,
	);
	
	return &(this->public);
}
