/*
 * Copyright (C) 2018 Tobias Brunner
 * Copyright (C) 2014 Martin Willi
 *
 * Copyright (C) secunet Security Networks AG
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

/*
 * Copyright (C) 2019-2022 Marvell 
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "swanctl.h"
#include "command.h"

#include <stdlib.h>
#include <unistd.h>

#include <library.h>
#include <selectors/sec_label.h>

/*
 * Described in header
 */
char *swanctl_dir;

#ifdef _WIN32
static bool get_install_dir(char *install_dir)
{
	char path[PATH_MAX], *pos;
	HMODULE module;

	module = GetModuleHandle(NULL);
	if (!module)
	{
		return FALSE;
	}
	if (!GetModuleFileName(module, path, sizeof(path)))
	{
		return FALSE;
	}

	pos = strrchr(path, '\\');
	if (!pos)
	{
		return FALSE;
	}
	*pos = 0;

	pos = strstr(path, "\\usr\\");
	if (!pos)
	{
		return FALSE;
	}
	*pos = 0;

	memset(install_dir, 0, sizeof(install_dir));
	strcpy(install_dir, path);

	return TRUE;
}
#endif

/*
 * Described in header
 */
settings_t *load_swanctl_conf(char *file)
{
	settings_t *cfg;
	char buf[PATH_MAX];
	char fname[PATH_MAX] = { 0 };

	if (!file)
	{
		if (!strlen(swanctl_dir))
		{
			free(swanctl_dir);
			swanctl_dir = strdup(getcwd(buf, sizeof(buf)));
		}
		file = buf;
		snprintf(buf, sizeof(buf), "%s%s%s", swanctl_dir,
				 DIRECTORY_SEPARATOR, SWANCTL_CONF);
	}

	memcpy(fname, file, strlen(file));
#ifdef _WIN32
	int len = 0;
	if (get_install_dir(fname))
	{
		len = strlen(fname);
		if(strstr(file, "/etc/swanctl\\swanctl.conf") != NULL)
		{
			strcpy((&fname[len]), file);
		}
		else
		{
			memset(fname, 0, PATH_MAX);
			memcpy(fname, file, strlen(file));
		}
	}
	else
	{
		memset(fname, 0, PATH_MAX);
		memcpy(fname, file, strlen(file));
	}
	for (int i = 0; i < strlen(fname); i++)
	{
		if(fname[i] == '/')
		{
			fname[i] = '\\';
		}
	}
	memset(file, 0, PATH_MAX);
	memcpy(file, fname, strlen(fname));
#endif
	//DBG1(DBG_CFG, "STAGING swanctl - load_swanctl_conf - file=%s", file);

	cfg = settings_create(file);
	if (!cfg)
	{
		fprintf(stderr, "parsing '%s' failed\n", file);
		return NULL;
	}
	free(swanctl_dir);
	swanctl_dir = path_dirname(file);
	return cfg;
}

/*
 * Described in header
 */
void print_label(const char *prefix, const char *value)
{
	sec_label_t *label;
	chunk_t encoding;

	if (value)
	{
		encoding = chunk_from_hex(chunk_from_str((char*)value), NULL);
		label = sec_label_from_encoding(encoding);
		if (label)
		{
			printf("%s%s\n", prefix, label->get_string(label));
			label->destroy(label);
		}
		chunk_free(&encoding);
	}
}

/**
 * Cleanup library atexit()
 */
static void cleanup()
{
	free(swanctl_dir);
	lib->processor->cancel(lib->processor);
	library_deinit();
}

/**
 * Library initialization and operation parsing
 */
int main(int argc, char *argv[])
{
	atexit(cleanup);
	if (!library_init(NULL, "swanctl"))
	{
		exit(SS_RC_LIBSTRONGSWAN_INTEGRITY);
	}
	if (lib->integrity &&
		!lib->integrity->check_file(lib->integrity, "swanctl", argv[0]))
	{
		fprintf(stderr, "integrity check of swanctl failed\n");
		exit(SS_RC_DAEMON_INTEGRITY);
	}
	if (!lib->plugins->load(lib->plugins,
			lib->settings->get_str(lib->settings, "swanctl.load", PLUGINS)))
	{
		exit(SS_RC_INITIALIZATION_FAILED);
	}

	swanctl_dir = strdup(getenv("SWANCTL_DIR") ?: SWANCTLDIR);

	dbg_default_set_level(0);
	lib->processor->set_threads(lib->processor, 4);
	dbg_default_set_level(1);

	return command_dispatch(argc, argv);
}
